package xyz.lychee.lagfixer.libs.adventure.util;

import java.util.Arrays;
import java.util.Collections;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.function.Function;
import java.util.function.IntFunction;
import xyz.lychee.lagfixer.libs.annotations.Contract;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;

public final class Index<K, V> {
   private final Map<K, V> keyToValue;
   private final Map<V, K> valueToKey;

   private Index(final Map<K, V> keyToValue, final Map<V, K> valueToKey) {
      this.keyToValue = keyToValue;
      this.valueToKey = valueToKey;
   }

   public static <K, V extends Enum<V>> @NotNull Index<K, V> create(final Class<V> type, final @NotNull Function<? super V, ? extends K> keyFunction) {
      return create(type, keyFunction, (Enum[])type.getEnumConstants());
   }

   @SafeVarargs
   public static <K, V extends Enum<V>> @NotNull Index<K, V> create(final Class<V> type, final @NotNull Function<? super V, ? extends K> keyFunction, final @NotNull V... values) {
      return create((Object[])values, (IntFunction)((length) -> {
         return new EnumMap(type);
      }), (Function)keyFunction);
   }

   @SafeVarargs
   public static <K, V> @NotNull Index<K, V> create(final @NotNull Function<? super V, ? extends K> keyFunction, final @NotNull V... values) {
      return create(values, HashMap::new, keyFunction);
   }

   public static <K, V> @NotNull Index<K, V> create(final @NotNull Function<? super V, ? extends K> keyFunction, final @NotNull List<V> constants) {
      return create(constants, HashMap::new, keyFunction);
   }

   private static <K, V> @NotNull Index<K, V> create(final V[] values, final IntFunction<Map<V, K>> valueToKeyFactory, final @NotNull Function<? super V, ? extends K> keyFunction) {
      return create(Arrays.asList(values), valueToKeyFactory, keyFunction);
   }

   private static <K, V> @NotNull Index<K, V> create(final List<V> values, final IntFunction<Map<V, K>> valueToKeyFactory, final @NotNull Function<? super V, ? extends K> keyFunction) {
      int length = values.size();
      Map<K, V> keyToValue = new HashMap(length);
      Map<V, K> valueToKey = (Map)valueToKeyFactory.apply(length);

      for(int i = 0; i < length; ++i) {
         V value = values.get(i);
         K key = keyFunction.apply(value);
         if (keyToValue.putIfAbsent(key, value) != null) {
            throw new IllegalStateException(String.format("Key %s already mapped to value %s", key, keyToValue.get(key)));
         }

         if (valueToKey.putIfAbsent(value, key) != null) {
            throw new IllegalStateException(String.format("Value %s already mapped to key %s", value, valueToKey.get(value)));
         }
      }

      return new Index(Collections.unmodifiableMap(keyToValue), Collections.unmodifiableMap(valueToKey));
   }

   public @NotNull Set<K> keys() {
      return Collections.unmodifiableSet(this.keyToValue.keySet());
   }

   public @Nullable K key(final @NotNull V value) {
      return this.valueToKey.get(value);
   }

   public @NotNull K keyOrThrow(final @NotNull V value) {
      K key = this.key(value);
      if (key == null) {
         throw new NoSuchElementException("There is no key for value " + value);
      } else {
         return key;
      }
   }

   @Contract("_, null -> null; _, !null -> !null")
   public K keyOr(final @NotNull V value, final @Nullable K defaultKey) {
      K key = this.key(value);
      return key == null ? defaultKey : key;
   }

   public @NotNull Set<V> values() {
      return Collections.unmodifiableSet(this.valueToKey.keySet());
   }

   public @Nullable V value(final @NotNull K key) {
      return this.keyToValue.get(key);
   }

   public @NotNull V valueOrThrow(final @NotNull K key) {
      V value = this.value(key);
      if (value == null) {
         throw new NoSuchElementException("There is no value for key " + key);
      } else {
         return value;
      }
   }

   @Contract("_, null -> null; _, !null -> !null")
   public V valueOr(final @NotNull K key, final @Nullable V defaultValue) {
      V value = this.value(key);
      return value == null ? defaultValue : value;
   }

   public @NotNull Map<K, V> keyToValue() {
      return Collections.unmodifiableMap(this.keyToValue);
   }

   public @NotNull Map<V, K> valueToKey() {
      return Collections.unmodifiableMap(this.valueToKey);
   }
}
